/*	motion.c
 *
 *	Detect changes in a video stream.
 *	Copyright 2000 by Jeroen Vreeken (pe1rxq@amsat.org)
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <linux/types.h>
#include <linux/videodev.h>

#include <jpeglib.h>

#define DEF_WIDTH	352
#define DEF_HEIGHT	288
#define DEF_QUALITY	50
#define DEF_CHANGES	1500

#define NORM_DEFAULT	0
#define IN_DEFAULT	8
#define VIDEO_DEVICE		"/dev/video0"

void usage (void)
{
	printf("motion Version "VERSION", Copyright 2000 Jeroen Vreeken\n");
	printf("usage:\n");
	printf("\n");
	printf("motion [-c changes] [-d device] [-q quality] [-s widthxheight] [-t target dir]\n");
	printf("\n");
	printf("-c changes\tthreshold for detecting a change, default: %d\n", DEF_CHANGES);
	printf("-d device\tvideo4linux capture device, default: "VIDEO_DEVICE"\n");
	printf("-q quality\tJPEG image quality, default: %d\n", DEF_QUALITY);
	printf("-s widthxheight\tPicture size, default: %dx%d\n", DEF_WIDTH, DEF_HEIGHT);
	printf("-t target dir\tdestination for snapshots.\n");
}

int get_image (int dev, char *image, int width, int height, int input, int norm,  int fmt)
{
        struct video_capability vid_caps;
	struct video_channel vid_chnl;
        struct video_window vid_win;


	if (ioctl (dev, VIDIOCGCAP, &vid_caps) == -1) {
		perror ("ioctl (VIDIOCGCAP)");
		return (-1);
	}
	/* Set channels and norms, NOT TESTED my philips cam doesn't have a
	 * tuner. */
        if (input != IN_DEFAULT) {
                vid_chnl.channel = -1;
                if (ioctl (dev, VIDIOCGCHAN, &vid_chnl) == -1) {
                        perror ("ioctl (VIDIOCGCHAN)");
                } else {
                        vid_chnl.channel = input;
                        vid_chnl.norm    = norm;
                        if (ioctl (dev, VIDIOCSCHAN, &vid_chnl) == -1) {
                                perror ("ioctl (VIDIOCSCHAN)");
                                return (-1);
                        }
                }
        }
	/* Set image size */
	if (ioctl (dev, VIDIOCGWIN, &vid_win) == -1)
		return (-1);
	vid_win.width=width;
	vid_win.height=height;
	if (ioctl (dev, VIDIOCSWIN, &vid_win) == -1)
		return (-1);
	/* Read an image */
	return read (dev, image, width * height * 3);
}

void put_image_jpeg (FILE *picture, char *image, int width, int height, int quality)
{
        int y, x, line_width;
        JSAMPROW row_ptr[1];
        struct jpeg_compress_struct cjpeg;
        struct jpeg_error_mgr jerr;
        char *line;

        line = malloc (width * 3);
        if (!line)
                return;
        cjpeg.err = jpeg_std_error(&jerr);
        jpeg_create_compress (&cjpeg);
        cjpeg.image_width = width;
        cjpeg.image_height= height;
        cjpeg.input_components = 3;
        cjpeg.in_color_space = JCS_RGB;
        jpeg_set_defaults (&cjpeg);

        jpeg_set_quality (&cjpeg, quality, TRUE);
        cjpeg.dct_method = JDCT_FASTEST;
        jpeg_stdio_dest (&cjpeg, picture);

        jpeg_start_compress (&cjpeg, TRUE);

        row_ptr[0] = line;
        line_width = width * 3;
	for ( y = 0; y < height; y++) {
        for (x = 0; x < line_width; x+=3) {
                        line[x]   = image[x+2];
                        line[x+1] = image[x+1];
                        line[x+2] = image[x];
                }
                jpeg_write_scanlines (&cjpeg, row_ptr, 1);
                image += line_width;
        }
        jpeg_finish_compress (&cjpeg);
        jpeg_destroy_compress (&cjpeg);
        free (line);
}


int main (int argc, char **argv)
{
	int width=DEF_WIDTH;
	int height=DEF_HEIGHT;
	int quality=DEF_QUALITY;
	int max_changes=DEF_CHANGES;
	int dev, i, diffs, shots=0;
	char *image1, *image2, *image_out;
	char *device=VIDEO_DEVICE;
	char file[255], filepath[255], c;
	FILE *picture;

	filepath[0]=0;
	while ((c=getopt(argc, argv, "c:d:s:q:t:h"))!=EOF) {
		switch (c) {
			case 'c':
				sscanf (optarg, "%d", &max_changes);
				break;
			case 'd':
				device=optarg;
				break;
			case 's':
				sscanf (optarg, "%dx%d", &width, &height);
				break;
			case 'q':
				sscanf (optarg, "%d", &quality);
				break;
			case 't':
				sscanf (optarg, "%s", filepath);
				break;
			case 'h':
			/* Do the same as an invallid option, but without the
			 * 'invalid option' error.... */
			default:
				usage();
				exit(-1);
		}
	}

	image1=malloc(width*height*3);
	image2=malloc(width*height*3);
	image_out=malloc(width*height*3);

	dev=open (device, O_RDWR);
	if (dev < 0) {
		perror ("Failed to open video device");
		exit(-1);
	}
	/* Capture first image, or we will get a alarm on start */
	get_image (dev, image1, width, height, IN_DEFAULT, NORM_DEFAULT, 
	    VIDEO_PALETTE_RGB24);

	/* Should go on forever... emphasis on 'should' */
	while (get_image (dev, image2, width, height, IN_DEFAULT, NORM_DEFAULT, 
	    VIDEO_PALETTE_RGB24)==width*height*3) {
		/* Make a differences picture in image_out */
		diffs=0;
		for (i=0; i<width*height*3; i++) {
			image_out[i]=image1[i]-image2[i];
			if (image_out[i] > 32 || image_out[i] < -32) {
				image_out[i]=image2[i];
				diffs++;
			} else {
				image_out[i]=0;
			}
		}
		/* The cat just walked in :) */
		if (diffs > max_changes) {
			printf("\007Motion detected %d\n", diffs);
			sprintf(file, "%04d-%ld.jpg", shots, time(NULL));
			printf("Picture saved to: %s\n", file);
			picture=fopen(file, "w");
			/* Output the latest picture 'image2'.
			 * replace it with 'image_out' for neat effects, but less
			 * recognizable
			 */
			put_image_jpeg (picture, image2, width, height, quality);
			fclose(picture);
			shots++;
		}
		/* Use this to minimize the number of pictures, also reduces
		 * the chance of a good one... However it can cause some
		 * seconds delay due to buffering in the v4l drivers...
		 */
//		sleep(1);
		memcpy(image1, image2, width*height*3);
	}
	perror("Somebody stole the video device, lets hope we got his picture");
	close (dev);
	free(image1);
	free(image2);
	free(image_out);
	exit(0);
}